import io
import cbor2
import dag_cbor


def read_event_pair(msg: bytes):
    with io.BytesIO(msg) as fp:
        header = cbor2.load(fp)
        event = dag_cbor.decode(fp.read())
    return header, event



import os
from functools import wraps

USING_NUMBA = False

try:
    assert 'NUMBA_DISABLE_JIT' not in os.environ
    from numba import jit
    USING_NUMBA = True
except (ImportError, AssertionError):
    def jit(*args, **kwargs):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator



import importlib
import os 


def test_jit_flags():
    import carbox.jit as jit

    orig_value = os.environ.get('NUMBA_DISABLE_JIT')
    assert jit.USING_NUMBA == (orig_value is None)

    if not orig_value:
        os.environ['NUMBA_DISABLE_JIT'] = '1'
        importlib.reload(jit)
        assert not jit.USING_NUMBA

        @jit.jit(nopython=True)
        def square(x: int) -> int:
            return x * x
        
        assert square(5) == 25

    if orig_value is None:
        os.environ.pop('NUMBA_DISABLE_JIT')
        importlib.reload(jit)




import io
import carbox.varint as varint


def test_read_uvarint():
    test_data = b"\x8E\x02\x03"  # Represents the uvarint values 142, 3

    buf = io.BufferedReader(io.BytesIO(test_data))

    buf_res_1 = varint.read_varint_from_reader(buf)
    assert buf_res_1 == 270

    buf_res_2 = varint.read_varint_from_reader(buf)
    assert buf_res_2 == 3

    bytes_res_1, offset1 = varint.read_varint_from_bytes(test_data)
    assert bytes_res_1 == buf_res_1

    bytes_res_2, offset2 = varint.read_varint_from_bytes(test_data, offset1)
    assert bytes_res_2 == buf_res_2
    assert offset2 == len(test_data)

    offset1 = varint.burn_varint_bytes(test_data, 0)
    offset2 = varint.burn_varint_bytes(test_data, offset1)
    assert offset2 == len(test_data)



from __future__ import annotations
import dag_cbor
import io
from typing import List, Tuple
from multiformats import CID
from carbox.jit import jit

from carbox.varint import (
    burn_varint_bytes,
    read_varint_from_bytes,
    read_varint_from_bytes_no_throw,
)


MAX_ALLOWED_SECTION_SIZE = 32 << 20  # 32MiB


class Block:
    __slots__ = ("_cid", "_data", "_decoded")

    def __init__(self, cid: CID, data: bytes):
        self._cid = cid
        self._data = data
        self._decoded = None

    @property
    def cid(self) -> CID:
        return self._cid

    @property
    def data(self) -> bytes:
        return self._data

    def __repr__(self) -> str:
        return f"Block(cid={self.cid}, data=[{len(self.data)} bytes])"

    def __eq__(self, other: Block) -> bool:
        """
        Compare blocks by CID only (assumes data was validated)
        """
        return self.cid == other.cid

    def __hash__(self) -> int:
        return hash(self.cid)

    @classmethod
    def from_bytes(cls, b: bytes, validate: bool = True) -> Block:
        cid_bytes, data = _read_cid_bytes_and_data(b, 0)
        cid = CID.decode(cid_bytes)
        if validate:
            got = cid.hashfun.digest(data)
            if got != cid.digest:
                raise ValueError(f"CID digest does not match: {cid.digest} != {got}")
        return cls(cid=cid, data=data)

    @property
    def decoded(self) -> dag_cbor.IPLDKind:
        value = self._decoded
        if value is None:
            value = dag_cbor.decode(self.data)
            self._decoded = value
        return value


def read_car(blocks: bytes, validate=True) -> Tuple[List[CID], List[Block]]:
    """
    Read a car file into a list of blocks.

    From the ipld CARv1 spec:

        |--------- Header --------| |---------- Data ------------|
        [ varint | DAG-CBOR block ] [ varint | CID | block ]{1,}

    :param blocks: the car bytes
    :param validate: if True, validate the roots against the blocks
    """
    roots, offset = _get_roots(blocks, 0)
    blocks = read_blocks(blocks, offset, validate)
    return roots, blocks


@jit(nopython=True)
def _read_cid_bytes_and_data(b: bytes, start_pos: int) -> Tuple[bytes, bytes]:
    # This borders on premature optimization. But since I
    # save a lot of firehose segments where I do validation up
    # before write and then never on read, this is useful to me.
    offset = start_pos
    for i in range(3):
        # Version, size, and hash func
        offset = burn_varint_bytes(b, offset)
        if offset == -1:
            raise ValueError(f"Invalid block format: uvarint loop {i}")

    # The digest size is variable
    n_digest_bytes, offset = read_varint_from_bytes_no_throw(b, offset)
    if n_digest_bytes == -1:
        raise ValueError("Invalid block format: digest size uvarint")

    end_pos = offset + n_digest_bytes
    if end_pos >= len(b):
        raise ValueError("Invalid block format: digest past end of data")

    return b[0:end_pos], b[end_pos:]


def read_blocks(data: bytes, offset: int, validate: bool) -> List[Block]:
    # Consume blocks
    blocks = []
    try:
        while True:
            block, offset = _ld_read_bytes(data, offset)
            blocks.append(Block.from_bytes(block, validate=validate))
    except EOFError:
        pass

    return blocks


def _get_roots(block_fp: io.BufferedReader, offset: int) -> List[CID]:
    header_buf, offset = _ld_read_bytes(block_fp, offset)
    if not header_buf:
        raise EOFError("Unexpected EOF while reading header")

    header = dag_cbor.decode(header_buf)
    if header["version"] != 1:
        raise ValueError(f"Unknown version {header.version}")

    if len(header["roots"]) == 0:
        raise ValueError("Empty car, no roots")

    return header["roots"], offset


@jit(nopython=True)
def _ld_read_bytes(b: bytes, offset: int) -> Tuple[bytes, int]:
    # See: https://github.com/ipld/go-car/blob/164c23bc8c01a3482a8be9c75631425239fda8c6/util/util.go#L113
    section_size, i = read_varint_from_bytes(b, offset)
    j = i + section_size
    if j > len(b):
        raise ValueError("Failed to read full buffer")

    if section_size > MAX_ALLOWED_SECTION_SIZE:  # Don't OOM
        raise ValueError(f"Section size {section_size} > {MAX_ALLOWED_SECTION_SIZE}")

    return b[i:j], j


class ParsedBlocks:
    __slots__ = ("_header", "_blocks")

    def __init__(self, blocks: bytes, validate: bool = True):
        self._header, self._blocks = read_car(blocks, validate=validate)

    @property
    def header(self) -> bytes:
        return self._header

    @property
    def blocks(self) -> List[Block]:
        return self._blocks

    def __getitem__(self, index: int) -> Block:
        return self._blocks[index]



from carbox.message import read_event_pair
from multiformats import CID
from base64 import b64decode
from carbox.car import read_car
from pathlib import Path


FIXTURES_DIR = Path(__file__).parent / "fixtures"

def test_read_car():
    fixture_path = FIXTURES_DIR / "donkeyballs.b64"
    post = b64decode(fixture_path.read_text())
    _, event = read_event_pair(post)
    blocks_raw = event["blocks"]
    roots, blocks = read_car(blocks_raw)

    assert [str(cid) for cid in roots] == [
        "zdpuArKcqh4Bfc5ufSWKTSS1jFRYJ47gpuxCEVXeWdMEjDpAM"
    ]
    assert len(blocks) == 10

    assert blocks[0].cid == CID.decode(
        "zdpuAx7GYAybGShxy9wvkK5eJt6a5G47tz5z5yeFcDqChfYE3"
    )
    assert blocks[0].decoded == {
        "$type": "app.bsky.feed.post",
        "createdAt": "2023-04-23T23:05:15.184Z",
        "text": "donkeyballs",
    }



import io
from typing import Tuple
from carbox.jit import jit


def read_varint_from_reader(reader: io.BufferedReader) -> int:
    shift = 0
    result = 0
    while True:
        b = reader.read(1)
        if not b:
            raise EOFError("Unexpected EOF")

        byte = b[0]
        result |= (byte & 0x7F) << shift
        if byte & 0x80 == 0:
            break
        shift += 7

    return result


@jit(nopython=True)
def read_varint_from_bytes(data: bytes, offset: int = 0) -> Tuple[int, int]:
    result, offset = read_varint_from_bytes_no_throw(data, offset)
    if result == -1:
        raise EOFError("Unexpected EOF while reading uvarint")
    return result, offset


@jit(nopython=True, inline="always")
def read_varint_from_bytes_no_throw(data: bytes, offset: int = 0) -> Tuple[int, int]:
    # TODO: guard against BigInt?
    shift = 0
    result = 0
    while True:
        if offset >= len(data):
            return -1, offset

        byte = data[offset]
        result |= (byte & 0x7F) << shift
        offset += 1

        if byte & 0x80 == 0:
            break
        shift += 7

    return result, offset


@jit(nopython=True, inline="always")
def burn_varint_bytes(data: bytes, offset: int = 0) -> int:
    while True:
        if offset >= len(data):
            return -1

        byte = data[offset]
        offset += 1

        if byte & 0x80 == 0:
            break

    return offset
